
/****************************************************************

    This program implements a simulated annealing for the
    quadratic assignment problem along the lines describes in
    the article D. T. Connoly, "An improved annealing scheme for 
    the QAP", European Journal of Operational Research 46, 1990,
    93-100. Based on sa_qap.cpp by E. Taillard,  EIVD, Route de 
    Cheseaux 1, CH-1400 Yverdon, Switzerland. Traqnslated from 
    C++ to Plan 9's C dialect by  Peter A. Cejchan 
    <cej@gli.cas.cz | pac7@post.cz | pac@plan9.cz>



    Format of data file : Example for problem nug5 :

5

0 1 1 2 3
1 0 2 1 2
1 2 0 1 2
2 1 1 0 1
3 2 2 1 0

0 5 2 4 1
5 0 3 0 2
2 3 0 0 0
4 0 0 0 5
1 2 0 5 0

   Additionnal parameters : Number of iterations, number of runs

 ********************************************************************/
//includes
#include <u.h>
#include <libc.h>
#include <bio.h>

//defines
#define ITER 10000
#define RUNS 10 
#define n_max 851
    
//constants
const long infini = 99999999;
const long nb_iter_initialisation = 1000; // Connolly proposes nb_iterations/100

//constants for random number generators
const long m = 2147483647; const long m2 = 2145483479; 
const long a12 = 63308; const long q12 = 33921; const long r12 = 12979; 
const long a13 = -183326; const long q13 = 11714; const long r13 = 2883; 
const long a21 = 86098; const long q21 = 24919; const long r21 = 7417; 
const long a23 = -539608; const long q23 = 3976; const long r23 = 2071;
const double invm = 4.656612873077393e-10;
long x10 = 12345, x11 = 67890, x12 = 13579, 
     x20 = 24680, x21 = 98765, x22 = 43210;

//typedefs
typedef long  type_vecteur[n_max];
typedef long type_matrice[n_max][n_max];

//declarations
long  n, nb_iterations, nb_res, no_res, Cout;
type_matrice a, b;
type_vecteur p;
Biobuf *bin;
enum booleen {faux, vrai};

//prototypes
void in_open(int argc, char **argv);
void in_close(Biobuf *bin);
int  mat_readdim(long *dim, Biobuf *bin);
int mat_read(type_matrice M, long rows, long cols, Biobuf *bin);
double max(double a, double b);
double min(double a, double b);
void swap(long *a, long *b);
double mon_rand(void);
long unif(long low, long high);
long calc_delta_complet2(long n, type_matrice a, type_matrice  b, type_vecteur  p, long r, long s);
long calcule_cout(long n, type_matrice  a, type_matrice b, type_vecteur p);
void tire_solution_aleatoire(long n, type_vecteur  p);
void recuit(long n, type_matrice a, type_matrice b, type_vecteur  meilleure_sol, long *meilleur_cout, long nb_iterations);


//functions i/o

int 
mat_readdim(long *dim, Biobuf *bin)
// reads #rows (columns) for a square matrix
{
	char *buf, *av[n_max], first;
	int n;

	n=0;
	while(n==0){
		if((buf = Brdstr(bin, '\n', 1)) == nil) return 0;			// nothing to read
		first=buf[0];
		n=tokenize(buf, av, 2);
		if(n!=0){
			// skip comment lines
			if((first=='#') || (first=='>')){
				n=0;
			}
		}
		else{
		// skip whitespace lines
		}
	}	// ends up with n > 0
	*dim=atoi(av[0]);
	return 1;
}

int
mat_read(type_matrice M, long rows, long cols, Biobuf *bin)
// reads in a whitespace and newline separated matrix
// CAUTION!! Indices start at 1, not 0, this is HIGHLY-NON-STANDARD, sorry, Eric!
{

	char *buf, *av[n_max];
	int i, j, n;
	for(i=1; i<=rows; i++){
		if((buf = Brdstr(bin, '\n', 1)) != nil){	
			// skip comment lines
			while((buf[0]=='#') || (buf[0]=='>')){
				if((buf = Brdstr(bin, '\n', 1)) == nil) return 0;	// read another line
			}	
			n=tokenize(buf, av, cols);
			// skip whitespace lines
			while(n==0){
				if((buf = Brdstr(bin, '\n', 1)) == nil) return 0;	// read another line
				n=tokenize(buf, av, cols);
			}	
			for(j=1; j<=cols; j++)
				M[i][j]=atol(av[j-1]);
			free(buf);
		}
		else return 0;
	}
	return 1;
}

void
in_open(int argc, char **argv)
//opens input file
{
	char *infile="/fd/0";
	if(argc < 2) {
		infile="/fd/0";
	}

		if(argc == 2) {

		infile=argv[1];
	}
	if ((bin = Bopen(infile, OREAD)) == 0) {
		exits("open");
	}
}

void
in_close(Biobuf *bin)
//closes input file
{
	Bterm(bin);
}

double max(double a, double b) {if (a > b) return(a); else return(b);}

double min(double a, double b) {if (a < b) return(a); else return(b);}

void swap(long *a, long *b) {long temp = *a; *a = *b; *b = temp;}

double mon_rand(void)
//random number generator
 {
	long h, p12, p13, p21, p23;
	h = x10/q13; p13 = -a13*(x10-h*q13)-h*r13;
	h = x11/q12; p12 = a12*(x11-h*q12)-h*r12;
	if (p13 < 0) p13 = p13 + m; if (p12 < 0) p12 = p12 + m;
	x10 = x11; x11 = x12; x12 = p12-p13; if (x12 < 0) x12 = x12 + m;
	h = x20/q23; p23 = -a23*(x20-h*q23)-h*r23;
	h = x22/q21; p21 = a21*(x22-h*q21)-h*r21;
	if (p23 < 0) p23 = p23 + m2; if (p21 < 0) p21 = p21 + m2;
	x20 = x21; x21 = x22; x22 = p21-p23; if(x22 < 0) x22 = x22 + m2;
	if (x12 < x22) h = x12 - x22 + m; else h = x12 - x22;
	if (h == 0) return(1.0); else return(h*invm);
 }

long unif(long low, long high)
//random number generator
 {
	return(low + (long)((high - low + 1.0) *  mon_rand() ));
 }


long calc_delta_complet2(long n, type_matrice a, type_matrice  b,
                         type_vecteur  p, long r, long s)
 {
	long d, k;
	d = (a[r][r]-a[s][s])*(b[p[s]][p[s]]-b[p[r]][p[r]]) +
		(a[r][s]-a[s][r])*(b[p[s]][p[r]]-b[p[r]][p[s]]);
	for (k = 1; k <= n; k = k + 1) 
		if (k!=r && k!=s)
			d = d + (a[k][r]-a[k][s])*(b[p[k]][p[s]]-b[p[k]][p[r]]) +
				(a[r][k]-a[s][k])*(b[p[s]][p[k]]-b[p[r]][p[k]]);
	return(d);
 }

long calcule_cout(long n, type_matrice  a, type_matrice b, type_vecteur p)
 {
	long i, j;
	long c = 0;
	for (i = 1; i <= n;  i++) 
		for (j = 1; j <= n; j++)
			c = c + a[i][j] * b[p[i]][p[j]];
	return(c);
 }

void tire_solution_aleatoire(long n, type_vecteur  p)
 {
	long i;
	for (i = 1; i <= n; i++) p[i] = i;
	for (i = 2; i <= n; i++) swap(&p[i-1], &p[unif(i-1, n)]);
 }

void recuit(long n, type_matrice a, type_matrice b,
	type_vecteur  meilleure_sol, long *meilleur_cout,
	long nb_iterations)
{
	type_vecteur p;
	long i, r, s;
	long delta, k, dmin,dmax,mxfail , nb_fail, no_iteration, Cout;
	double t0, tf, beta, tfound, temperature;

	k = n*(n-1)/2;
	mxfail = k;
	dmin = infini, 
	dmax = 0;

	for (i = 1; i <= n; i++) p[i] = meilleure_sol[i];
	Cout = calcule_cout(n, a, b, p);
	*meilleur_cout = Cout;

	for (no_iteration = 1; no_iteration <= nb_iter_initialisation; no_iteration++) {
		r = unif(1, n);
		s = unif(1, n-1);
		if (s >= r) s = s+1;
		delta = calc_delta_complet2(n,a,b,p,r,s);
		if (delta > 0) {
			dmin = min(dmin, delta); dmax = max(dmax, delta);
		}; 
		Cout = Cout + delta;
		swap(&p[r], &p[s]);
	};
	t0 = dmin + (dmax - dmin)/10.0;
	tf = dmin;
	beta = (t0 - tf)/(nb_iterations*t0*tf);

	nb_fail = 0;
	tfound = t0;
	temperature = t0;
	r = 1; s = 2;
	for (no_iteration = 1; no_iteration <= nb_iterations - nb_iter_initialisation; no_iteration++) {
		temperature = temperature / (1.0 + beta*temperature);
		s = s + 1;
		if (s > n) {
			r = r + 1; 
			if (r > n - 1) r = 1;
			s = r + 1;
		};

		delta = calc_delta_complet2(n,a,b,p,r,s);
		if ((delta < 0) || (mon_rand() < exp(- (double)(delta)/temperature)) || mxfail == nb_fail) {
			Cout = Cout + delta; swap(&p[r], &p[s]); nb_fail = 0;
		}
		else nb_fail = nb_fail + 1;

		if (mxfail == nb_fail) {beta = 0; temperature = tfound;};
		if (Cout < *meilleur_cout) {
			*meilleur_cout = Cout;
			for (i = 1; i <= n; i++) meilleure_sol[i] = p[i];
			tfound = temperature;
			print("Iteration = %ld ", no_iteration);
			print("Cost = %ld\n" , *meilleur_cout);
		};
	};

	print("Best solution found : \n");
	for (i = 1; i <= n; i = i + 1) 
		print("%ld ", meilleure_sol[i]);
	print("\n");
	
 }


void
main(int argc, char **argv)
{
	in_open(argc, argv);
	mat_readdim(&n, bin);
	mat_read(a, n, n, bin);
	mat_read(b, n, n, bin);
	in_close(bin);
	nb_iterations = ITER; nb_res = RUNS;
	for (no_res = 1; no_res <= nb_res; no_res = no_res + 1) {
		tire_solution_aleatoire(n, p);
		 recuit(n,a,b,p,&Cout, nb_iterations);
	};
 }
//eof
